from motorway.discovery.base import DiscoveryBase


class LocalDiscovery(DiscoveryBase):
    def get_queue_uri(self, queue, direction):
        assert direction in ('push', 'pull')
        # return "ipc://livestats_motorway_%s-%s" % (queue, direction)
        return "ipc://livestats_motorway_%s" % (queue, )


from motorway.grouping import HashRingGrouper, RandomGrouper


class GrouperMixin(object):
    def get_grouper(self, grouper_name):
        if grouper_name == "HashRingGrouper":
            return HashRingGrouper
        else:
            return RandomGrouper


from multiprocessing import Process
from setproctitle import setproctitle
import time
from motorway.controller import ControllerIntersection
from motorway.discovery.local import LocalDiscovery
from motorway.utils import ramp_result_stream_name
import zmq
import logging
from motorway.webserver import WebserverIntersection

logger = logging.getLogger(__name__)


class Pipeline(object):
    def __init__(self, discovery_backend=LocalDiscovery, controller_bind_address="0.0.0.0:7007"):
        self._streams = {}
        self._stream_consumers = {}
        self._processes = []
        self._queue_processes = []
        self._ramp_result_streams = []
        self._discovery_backend = discovery_backend()
        self.controller_bind_address = "tcp://%s" % controller_bind_address
        self.context = zmq.Context()

    def definition(self):
        """
        Extend this method in your :class:`motorway.pipeline.Pipeline` subclass, e.g.::

            class WordCountPipeline(Pipeline):
                def definition(self):
                    self.add_ramp(WordRamp, 'sentence')
                    self.add_intersection(SentenceSplitIntersection, 'sentence', 'word', processes=2)
                    self.add_intersection(WordCountIntersection, 'word', 'word_count', grouper_cls=HashRingGrouper, processes=2)
                    self.add_intersection(AggregateIntersection, 'word_count', grouper_cls=HashRingGrouper, processes=1)
        """
        raise NotImplementedError("You must implement a definition() on your pipeline")

    def _add_process(self, cls, process_instances, process_args, input_stream=None, output_stream=None, show_in_ui=True, process_start_number=0):
        for i in range(process_start_number, process_instances + process_start_number):
            process_name = "%s-%s" % (cls.__name__, i)
            p = Process(
                target=cls.run,
                args=process_args,
                name=process_name
            )
            self._processes.append(p)
            if show_in_ui:
                if output_stream:
                    if output_stream not in self._stream_consumers:
                        self._stream_consumers[output_stream] = {'producers': [], 'consumers': []}
                    self._stream_consumers[output_stream]['producers'].append(process_name)
                if input_stream not in self._stream_consumers:
                    self._stream_consumers[input_stream] = {'producers': [], 'consumers': []}
                self._stream_consumers[input_stream]['consumers'].append(process_name)

    def add_ramp(self, ramp_class, output_stream, processes=1):
        ramp_result_stream = ramp_result_stream_name(ramp_class.__name__)
        self._ramp_result_streams.append((ramp_class.__name__, ramp_result_stream))
        self._add_process(
            ramp_class,
            processes,
            process_args=(
                output_stream,
                self.controller_bind_address
            ),
            output_stream=output_stream,

        )

    def add_intersection(self, intersection_class, input_stream, output_stream=None, processes=1, grouper_cls=None):
        self._add_process(
            intersection_class,
            processes,
            process_args=(
                input_stream,
                output_stream,
                self.controller_bind_address,
                grouper_cls
            ),
            input_stream=input_stream,
            output_stream=output_stream,
        )

    def _add_controller(self):
        self._add_process(
            ControllerIntersection,
            1,
            process_args=(
                self.controller_bind_address,
                self._stream_consumers,
            ),
            process_start_number=1
        )

    def run(self):
        """
        Execute the entire pipeline in several sub processes.

        """

        logger.info("Starting Pipeline!")

        setproctitle("data-pipeline: main")

        # User jobs
        self.definition()
        logger.debug("Loaded definition")

        # Controller Transformer
        # self._add_controller()
        self.add_intersection(ControllerIntersection, None, '_web_server')
        self.add_intersection(WebserverIntersection, '_web_server')

        logger.debug("Running queues")
        for queue_process in self._queue_processes:
            queue_process.start()
        logger.debug("Running pipeline")
        for process in self._processes:
            process.start()

        self._processes += self._queue_processes  # Monitor queue processes and shut then down with the rest
        try:
            while True:
                for process in self._processes:
                    assert process.is_alive()
                time.sleep(0.5)
        except Exception:
            raise
        finally:
            self.kill()
        logger.debug("Finished Pipeline!")

    def kill(self):
        for process in self._processes:
            logger.warn("Terminating %s" % process)
            if process.is_alive():
                process.terminate()



import datetime
import json
from threading import Thread
from flask import Flask, render_template, Response
from motorway.intersection import Intersection
from motorway.utils import DateTimeAwareJsonEncoder


class WebserverIntersection(Intersection):
    def __init__(self):
        super(WebserverIntersection, self).__init__()

        self.process_id_to_name = {}
        self.process_statistics = {}
        self.stream_consumers = {}
        self.failed_messages = {}

        app = Flask(__name__)

        @app.route("/")
        def index():
            return render_template("index.html")

        @app.route("/detail/<process>/")
        def detail(process):
            return render_template("detail.html", process=process)

        @app.route("/api/status/")
        def api_status():
            now = datetime.datetime.now()
            return Response(json.dumps(dict(
                sorted_process_statistics=sorted(
                    [(self.process_id_to_name.get(process_id, process_id), stats) for process_id, stats in self.process_statistics.items()],
                    key=lambda itm: itm[0]
                ),
                stream_consumers=self.stream_consumers,
                last_minutes=[(now - datetime.timedelta(minutes=i)).minute for i in range(0, 10)]
            ), cls=DateTimeAwareJsonEncoder), mimetype='application/json')

        @app.route("/api/detail/<process>/")
        def api_detail(process):
            return Response(json.dumps(dict(
                failed_messages=[msg[1] for msg in self.failed_messages.values() if self.process_id_to_name[msg[0]] == process],
            ), cls=DateTimeAwareJsonEncoder), mimetype='application/json')

        p = Thread(target=app.run, name="motorway-webserver", kwargs=dict(
            port=5000,
            host="0.0.0.0",
        ))
        p.start()

    def process(self, message):
        self.process_id_to_name = message.content['process_id_to_name']
        self.process_statistics = message.content['process_statistics']
        self.stream_consumers = message.content['stream_consumers']
        self.failed_messages = message.content['failed_messages']
        yield


from collections import Counter
import json
import logging
from setproctitle import setproctitle
from threading import Thread
import time
import datetime
import uuid
from motorway.decorators import batch_process
from motorway.messages import Message
from motorway.intersection import Intersection
from motorway.utils import ramp_result_stream_name, percentile_from_dict, DateTimeAwareJsonEncoder, \
    set_timeouts_on_socket
from flask import Flask, render_template, Response
from isodate import parse_duration
import zmq


logger = logging.getLogger(__name__)


def resolve_tree(stream_consumers, producer):
    streams = {}
    for stream, dct in stream_consumers.items():
        if producer in dct['producers']:
            intersections = {}
            for consumer in dct['consumers']:
                intersections[consumer] = resolve_tree(stream_consumers, consumer)
            streams[stream] = intersections
    return streams


class ControllerIntersection(Intersection):

    def get_default_process_dict(self):
        return {
            'success': 0,
            'failed': 0,
            'processed': 0,
            'waiting': 0,
            'time_taken': datetime.timedelta(seconds=0),
            'avg_time_taken': datetime.timedelta(seconds=0),
            '95_percentile': datetime.timedelta(seconds=0),
            'frequency': {}.copy(),
            'histogram': {minute: {'error_count': 0, 'success_count': 0, 'timeout_count': 0}.copy() for minute in range(0, 60)}.copy()
        }.copy()

    def __init__(self, stream_consumers, context, controller_bind_address, web_server=True):
        super(ControllerIntersection, self).__init__()
        self.stream_consumers = stream_consumers
        self.ramp_socks = {}
        self.messages = {}
        self.failed_messages = {}
        self.process_statistics = {}
        self.waiting_messages = {}
        self.queue_processes = {}
        self.context = context
        self.controller_bind_address = controller_bind_address
        self.process_id_to_name = {}  # Maps UUIDs to human readable names
        self.process_address_to_uuid = {}  # Maps tcp endpoints to human readable names


    @batch_process(wait=1, limit=500)
    def process(self, messages):
        for message in messages:
            original_process = message.producer_uuid
            if original_process not in self.process_statistics:
                self.process_statistics[original_process] = self.get_default_process_dict()

            current_process = message.content['process_name']
            if current_process not in self.process_statistics:
                self.process_statistics[current_process] = self.get_default_process_dict()

            destination_process = message.destination_uuid or self.process_address_to_uuid[message.destination_endpoint]
            if destination_process not in self.process_statistics:
                self.process_statistics[destination_process] = self.get_default_process_dict()

            # Create message or update with ack value
            if message.ramp_unique_id not in self.messages:  # Message just created
                assert message.producer_uuid, "Producer UUID missing from %s" % message._message()
                self.messages[message.ramp_unique_id] = [
                    message.producer_uuid,
                    message.ack_value,
                    datetime.datetime.now(),
                    destination_process
                ]  # Set the new value to ack value
            elif message.ack_value >= 0:  # Message processed
                self.messages[message.ramp_unique_id][1] ^= message.ack_value  # XOR the existing value
                # Update process information
                if self.messages[message.ramp_unique_id][1] == Message.SUCCESS:
                    original_process = self.messages[message.ramp_unique_id][0]
                    self.process_statistics[original_process]['success'] += 1
                    self.process_statistics[original_process]['histogram'][datetime.datetime.now().minute]['success_count'] += 1
                    self.success(message.ramp_unique_id, original_process)
                    del self.messages[message.ramp_unique_id]
                else:  # Still not finished - update destination!
                    if destination_process != original_process:
                        self.messages[message.ramp_unique_id][3] = destination_process  # Update destination so we can show what is waiting for this process!
            elif message.ack_value == Message.FAIL:
                if message.ramp_unique_id in self.messages:
                    original_process, ack_value, start_time, destination_process = self.messages[message.ramp_unique_id]
                    del self.messages[message.ramp_unique_id]
                self.process_statistics[original_process]['failed'] += 1
                self.process_statistics[original_process]['histogram'][datetime.datetime.now().minute]['error_count'] += 1
                self.fail(message.ramp_unique_id, error_message=message.error_message, process=original_process)
            self.process_statistics[original_process]['processed'] += 1

            # Update statistics
            if 'duration' in message.content:
                time_taken = parse_duration(message.content['duration'])
                rounded_seconds = round(time_taken.total_seconds(), 0)
                self.process_statistics[current_process]['time_taken'] += time_taken
                self.process_statistics[current_process]['frequency'][rounded_seconds] = self.process_statistics[current_process]['frequency'].get(rounded_seconds, 0) + 1
                self.process_statistics[current_process]['95_percentile'] = datetime.timedelta(seconds=percentile_from_dict(self.process_statistics[current_process]['frequency'], 95))
                self.process_statistics[current_process]['avg_time_taken'] = self.process_statistics[current_process]['time_taken'] / sum(self.process_statistics[current_process]['frequency'].values())
        yield  # Hack: This is actually done by self.update() to trigger it even if there are no messages and to reduce messages to 1/s

    def update(self):
        now = datetime.datetime.now()
        # Check message status
        waiting_messages = {}
        for unique_id, lst in self.messages.items():
            original_process, ack_value, start_time, process = lst
            if unique_id in self.failed_messages:
                del self.messages[unique_id]  # This failed somewhere else in the chain and it was notificed already
            elif (now - start_time) > datetime.timedelta(minutes=30):
                del self.messages[unique_id] # clean up
                self.process_statistics[process]['histogram'][datetime.datetime.now().minute]['timeout_count'] += 1
                self.fail(unique_id, original_process, error_message="Message timed out")
            elif ack_value > 0:
                waiting_messages[process] = waiting_messages.get(process, 0) + 1
        self.waiting_messages = waiting_messages

        # Update histograms
        for process in self.process_statistics.keys():
            self.process_statistics[process]['histogram'][(now + datetime.timedelta(minutes=1)).minute] = self.get_default_process_dict()['histogram'][0]  # reset next minute
            self.process_statistics[process]['waiting'] = self.waiting_messages.get(process, 0)

        message = Message("_controller-%s" % uuid.uuid4(), {
            'process_id_to_name': self.process_id_to_name,
            'process_statistics': self.process_statistics,
            'stream_consumers': self.stream_consumers,
            'failed_messages': self.failed_messages,
        })

        if self.send_socks:  # Send updates to webserves. Uses the same code as intersection._process - consider deduplicating it
            socket_address = self.get_grouper(self.send_grouper)(
                self.send_socks.keys()
            ).get_destination_for(message.grouping_value)
            message.send(
                self.send_socks[socket_address],
                self.process_uuid
            )

    def _update_wrapper(self):
        while True:
            self.update()
            time.sleep(1)

    def fail(self, unique_id, process, error_message=""):
        self.failed_messages[unique_id] = (process, error_message)
        if process not in self.ramp_socks:
            logging.debug("%s not in ramp_socks, probably a intersection which doesn't support feedback. Had %s" % (process, self.ramp_socks))
        else:
            self.ramp_socks[process].send_json({
                'status': 'fail',
                'id': unique_id
            })

    def success(self, unique_id, process):
        if process not in self.ramp_socks:
            logging.debug("%s not in ramp_socks, probably a intersection which doesn't support feedback. Had %s" % (process, self.ramp_socks))
        else:
            self.ramp_socks[process].send_json({
                'status': 'success',
                'id': unique_id
            })

    def update_connections(self, output_queue):
        update_connection_sock = self.context.socket(zmq.PULL)
        update_connection_port = update_connection_sock.bind_to_random_port("tcp://*")
        set_timeouts_on_socket(update_connection_sock)

        refresh_connection_sock = self.context.socket(zmq.PUB)
        refresh_connection_sock.bind(self.controller_bind_address)
        set_timeouts_on_socket(refresh_connection_sock)

        self.queue_processes['_update_connections'] = {
            'streams': ['tcp://127.0.0.1:%d' % update_connection_port],
            'grouping': None
        }
        refresh_connection_sock.send_json(self.queue_processes)  # Initial refresh

        poller = zmq.Poller()
        poller.register(update_connection_sock, zmq.POLLIN)

        while True:
            socks = dict(poller.poll(timeout=10000))
            if socks.get(update_connection_sock) == zmq.POLLIN:
                connection_updates = update_connection_sock.recv_json()
                self.process_id_to_name[connection_updates['meta']['id']] = connection_updates['meta']['name']
                for queue, consumers in connection_updates['streams'].items():
                    if queue not in self.queue_processes:
                        self.queue_processes[queue] = {
                            'streams': [],
                            'grouping': None
                        }
                    for consumer in consumers:
                        if consumer not in self.queue_processes[queue]['streams']:
                            self.process_address_to_uuid[consumer] = connection_updates['meta']['id']
                            # Add to streams and update grouping TODO: Keep track of multiple groupings
                            self.queue_processes[queue]['streams'].append(consumer)
                            self.queue_processes[queue]['grouping'] = connection_updates['meta']['grouping']
                            if '_ramp' in queue:  # Ramp replies
                                self.ramp_socks[queue] = self.context.socket(zmq.PUSH)
                                self.ramp_socks[queue].connect(consumer)
            refresh_connection_sock.send_json(self.queue_processes)
            logger.debug("Announced %s", self.queue_processes)
            # Update our own send socks
            if output_queue and output_queue in self.queue_processes:
                self.set_send_socks(self.queue_processes, output_queue, self.context)

    def _process_wrapper(self):
        message_ack_sock = self.context.socket(zmq.PULL)
        message_ack_port = message_ack_sock.bind_to_random_port("tcp://*")
        self.queue_processes['_message_ack'] = {
            'streams': ['tcp://127.0.0.1:%s' % message_ack_port],
            'grouping': None
        }
        set_timeouts_on_socket(message_ack_sock)
        while True:
            self._process(message_ack_sock, None)


    @classmethod
    def run(cls, input_stream, output_stream=None, refresh_connection_stream=None, grouper_cls=None):

        setproctitle("data-pipeline: %s" % cls.__name__)
        logger.info("Running %s" % cls.__name__)

        context = zmq.Context()

        self = cls(
            {},
            context,
            refresh_connection_stream
        )

        # Create Thread Factories :-)

        thread_update_connections_factory = lambda: Thread(target=self.update_connections, name="controller-update_connections", args=(output_stream, ))
        thread_update_stats_factory = lambda: Thread(target=self._update_wrapper, name="controller-update_stats")
        thread_process_factory = lambda: Thread(target=self._process_wrapper, name="controller-process_acks")

        # Run threads

        thread_update_connections = thread_update_connections_factory()
        thread_update_connections.start()

        thread_update_stats = thread_update_stats_factory()
        thread_update_stats.start()

        thread_process = thread_process_factory()
        thread_process.start()

        while True:
            if not thread_update_connections.isAlive():
                logger.error("Thread thread_update_connections crashed in %s" % self.__class__.__name__)
                thread_update_connections = thread_update_connections_factory()
                thread_update_connections.start()
            if not thread_update_stats.isAlive():
                logger.error("Thread thread_update_stats crashed in %s" % self.__class__.__name__)
                thread_update_stats = thread_update_stats_factory()
                thread_update_stats.start()
            if not thread_process.isAlive():
                logger.error("Thread thread_process crashed in %s" % self.__class__.__name__)
                thread_process = thread_process_factory()
                thread_process.start()
            time.sleep(10)



from Queue import Empty
import logging
import multiprocessing
import random
from threading import Thread
import time
from time import time as _time
import uuid
import datetime
from motorway.messages import Message
from motorway.mixins import GrouperMixin
from motorway.utils import set_timeouts_on_socket, get_connections_block
import zmq
from setproctitle import setproctitle

logger = logging.getLogger(__name__)


class Intersection(GrouperMixin, object):

    fields = []

    def __init__(self):
        self.messages_processed = 0
        self.process_uuid = str(uuid.uuid4())
        self.receive_port = None
        self.send_socks = {}
        self.send_grouper = None
        self.controller_sock = None

    def _process(self, receive_sock, controller_sock=None):

        try:
            if getattr(self.process, 'batch_process', None):
                poller = zmq.Poller()
                poller.register(receive_sock, zmq.POLLIN)
                value = []
                end_time = _time() + self.process.wait
                while end_time > _time() and len(value) < self.process.limit:
                    socks = dict(poller.poll(timeout=1000))
                    if socks.get(receive_sock) == zmq.POLLIN:
                        value.append(receive_sock.recv_json())
                    else:
                        time.sleep(1)
                self.messages_processed += len(value)
            else:
                poller = zmq.Poller()
                poller.register(receive_sock, zmq.POLLIN)
                socks = dict(poller.poll(timeout=1000))
                if socks.get(receive_sock) == zmq.POLLIN:
                    value = receive_sock.recv_json()
                else:
                    return
                self.messages_processed += 1
            if value:
                if isinstance(value, list):
                    message = [Message.from_message(m, controller_sock, process_name=self.process_uuid) for m in value]
                else:
                    message = Message.from_message(value, controller_sock, process_name=self.process_uuid)
                try:
                    for generated_message in self.process(message):
                        if generated_message is not None and self.send_socks:
                            socket_address = self.get_grouper(self.send_grouper)(
                                self.send_socks.keys()
                            ).get_destination_for(generated_message.grouping_value)
                            generated_message.send(
                                self.send_socks[socket_address],
                                self.process_uuid
                            )
                            if controller_sock:
                                generated_message.send_control_message(
                                    controller_sock,
                                    process_name=self.process_uuid,
                                    destination_endpoint=socket_address
                                )
                except Exception as e:
                    logger.error(str(e), exc_info=True)
                    if isinstance(message, list):
                        [m.fail() for m in message]
                    else:
                        message.fail()

        except Empty:  # Didn't receive anything from ZMQ
            pass

    def process(self, message):
        """
        This function is called continuously by the intersection.

        :yield: :class:`motorway.messages.Message` instance
        :param message: :class:`motorway.messages.Message` instance or :func:`list` if using
            :func:`motorway.decorators.batch_process`

        """
        raise NotImplementedError()

    @classmethod
    def run(cls, input_stream, output_stream=None, refresh_connection_stream=None, grouper_cls=None):
        self = cls()
        process_name = multiprocessing.current_process().name
        logger.info("Running %s" % process_name)
        logger.debug("%s listening on %s and pushing to %s" % (process_name, input_stream, output_stream))
        setproctitle("data-pipeline: %s" % process_name)
        context = zmq.Context()

        # Create Thread Factories :-)

        thread_update_connections_factory = lambda: Thread(target=self.connection_thread, name="connection_thread", kwargs={
            'refresh_connection_stream': refresh_connection_stream,
            'context': context,
            'input_queue': input_stream,
            'output_queue': output_stream,
            'process_id': self.process_uuid,
            'process_name': process_name,
            'grouper_cls': grouper_cls
        })

        thread_main_factory = lambda: Thread(target=self.receive_messages, name="message_producer", kwargs={
            'context': context,
            'output_stream': output_stream,
            'grouper_cls': grouper_cls,
        })

        # Run threads

        thread_update_connections = thread_update_connections_factory()
        thread_update_connections.start()

        thread_main = thread_main_factory()
        thread_main.start()

        while True:
            if not thread_update_connections.isAlive():
                logger.error("Thread thread_update_connections crashed in %s" % self.__class__.__name__)
                thread_update_connections = thread_update_connections_factory()
                thread_update_connections.start()
            if not thread_main.isAlive():
                logger.error("Thread thread_main crashed in %s" % self.__class__.__name__)
                thread_main = thread_main_factory()
                thread_main.start()
            time.sleep(10)

    def connection_thread(self, context=None, refresh_connection_stream=None, input_queue=None, output_queue=None, process_id=None,
                          process_name=None, grouper_cls=None):
        refresh_connection_sock = context.socket(zmq.SUB)
        refresh_connection_sock.connect(refresh_connection_stream)
        refresh_connection_sock.setsockopt(zmq.SUBSCRIBE, '')  # You must subscribe to something, so this means *all*}
        set_timeouts_on_socket(refresh_connection_sock)

        connections = get_connections_block('_update_connections', refresh_connection_sock)

        while not self.receive_port:
            time.sleep(1)
        # Register as consumer of input stream
        update_connection_sock = context.socket(zmq.PUSH)
        update_connection_sock.connect(connections['_update_connections']['streams'][0])
        update_connection_sock.send_json({
            'streams': {
                input_queue: ['tcp://127.0.0.1:%s' % self.receive_port]
            },
            'meta': {
                'id': process_id,
                'name': process_name,
                'grouping': None if not grouper_cls else grouper_cls.__name__
            }
        })

        connections = get_connections_block('_message_ack', refresh_connection_sock, existing_connections=connections)
        self.controller_sock = context.socket(zmq.PUSH)
        self.controller_sock.connect(connections['_message_ack']['streams'][0])
        set_timeouts_on_socket(self.controller_sock)

        while True:
            try:
                connections = refresh_connection_sock.recv_json()
                if output_queue in connections:
                    self.set_send_socks(connections, output_queue, context)

            except zmq.Again:
                pass

    def set_send_socks(self, connections, output_queue, context):
        for send_conn in connections[output_queue]['streams']:
            if send_conn not in self.send_socks:
                send_sock = context.socket(zmq.PUSH)
                send_sock.connect(send_conn)
                self.send_socks[send_conn] = send_sock
                self.send_grouper = connections[output_queue]['grouping']

    def receive_messages(self, context=None, output_stream=None, grouper_cls=None):
        receive_sock = context.socket(zmq.PULL)
        self.receive_port = receive_sock.bind_to_random_port("tcp://*")
        set_timeouts_on_socket(receive_sock)

        while not self.controller_sock:
            time.sleep(1)
        if output_stream:
            while not self.controller_sock:
                time.sleep(1)

        while True:
            self._process(receive_sock, controller_sock=self.controller_sock)





import random
from hash_ring import HashRing
import zmq
import logging

logger = logging.getLogger(__name__)


class GroupingValueMissing(Exception):
    pass


class HashRingGrouper(object):
    def __init__(self, destinations):
        self.hash_ring = HashRing(destinations)

    def get_destination_for(self, value):
        if type(value) is int:
            value = str(value)
        try:
            return self.hash_ring.get_node(value)
        except TypeError:
            raise GroupingValueMissing("'%s' is an invalid grouping value for HashRingGrouper" % (value, ))


class RandomGrouper(object):
    def __init__(self, destinations):
        self.destinations = destinations

    def get_destination_for(self, value):
        return random.choice(self.destinations)


class GroupingBuffer(object):
    @classmethod
    def run(cls, input_stream, output_stream_function, number_of_outputs, grouper_cls):
        logger.info("Running grouper %s" % input_stream)
        context = zmq.Context()

        receive_sock = context.socket(zmq.PULL)
        receive_sock.connect(input_stream)

        grouper = grouper_cls(number_of_outputs)

        processes = {}

        for process in range(0, number_of_outputs):
            send_sock = context.socket(zmq.PUSH)
            logger.debug("Grouper listening to %s and pushing to %s" % (input_stream, output_stream_function(process)))
            send_sock.connect(output_stream_function(process))
            processes[process] = send_sock

        while True:
            poller = zmq.Poller()
            poller.register(receive_sock, zmq.POLLIN)
            socks = dict(poller.poll(timeout=1000))
            if socks.get(receive_sock) == zmq.POLLIN:
                value = receive_sock.recv_json()
                destination = grouper.get_destination_for(str(value['grouping_value']))
                processes[destination].send_json(value)


from collections import defaultdict
import random
import time
import uuid
from motorway.contrib.amazon_sqs.intersections import SQSInsertIntersection
from motorway.decorators import batch_process
from motorway.messages import Message
from motorway.intersection import Intersection


class SentenceSplitIntersection(Intersection):
    def process(self, message):
        for word in message.content.split(" "):
            yield Message.new(message, word, grouping_value=word)
        message.ack()


class WordCountIntersection(Intersection):
    def __init__(self):
        self._count = defaultdict(int)
        super(WordCountIntersection, self).__init__()

    @batch_process(wait=2, limit=500)
    def process(self, messages):
        for message in messages:
            self._count[message.content] += 1
            message.ack()
        yield Message(uuid.uuid4(), self._count)


class AggregateIntersection(Intersection):
    def __init__(self):
        self._count = defaultdict(int)
        super(AggregateIntersection, self).__init__()

    def process(self, message):
        self._count.update(message.content)
        yield
        message.ack()
        print self._count


class ExampleSQSIntersection(SQSInsertIntersection):
    queue_name = "tutorial_motorway"


import logging
import multiprocessing
from setproctitle import setproctitle
import datetime
from threading import Thread
import uuid
from motorway.grouping import GroupingValueMissing
from motorway.mixins import GrouperMixin
from motorway.utils import set_timeouts_on_socket, get_connections_block
import zmq
import time
import random


logger = logging.getLogger(__name__)


class Ramp(GrouperMixin, object):

    fields = []

    def __init__(self):
        super(Ramp, self).__init__()
        self.send_socks = {}
        self.send_grouper = None
        self.result_port = None

    def next(self):
        """
        This function is called continuously by the ramp.

        .. WARNING::
            Do not block this for a long time or make a while True loop inside it. Betwen every :func:`motorway.ramp.Ramp.next` run,
            some essential operations are run, including receiving acks from the :class:`motorway.controller.Controller`

        :yield: :class:`motorway.messages.Message` instance
        """
        raise NotImplementedError()

    def __iter__(self):
        yield self.next()

    def success(self, _id):
        """
        Called when a message was successfully ack'ed through the entire pipeline.

        :param _id: The id of the message that was successful
        """
        pass

    def failed(self, _id):
        """
        Called when a message failed somewhere in the pipeline. The message might not be entirely finished processing
        at this point and this function might be called multiple times.

        :param _id: The id of the message that failed
        """
        pass

    @classmethod
    def run(cls, queue, refresh_connection_stream=None):

        self = cls()

        process_name = multiprocessing.current_process().name
        process_uuid = str(uuid.uuid4())
        process_id = "_ramp-%s" % process_uuid
        logger.info("Running %s" % process_name)
        logger.debug("%s pushing to %s" % (process_name, queue))
        setproctitle("data-pipeline: %s" % process_name, )

        context = zmq.Context()

        # Create Thread Factories :-)

        thread_update_connections_factory = lambda: Thread(target=self.connection_thread, name="connection_thread", kwargs={
            'refresh_connection_stream': refresh_connection_stream,
            'context': context,
            'queue': queue,
            'process_id': process_id,
            'process_name': process_name,
        })

        thread_main_factory = lambda: Thread(target=self.message_producer, name="message_producer", kwargs={
            'context': context,
            'process_id': process_id,
        })

        thread_results_factory = lambda: Thread(target=self.receive_replies, name="results", kwargs={
            'context': context,
        })

        # Run threads

        thread_update_connections = thread_update_connections_factory()
        thread_update_connections.start()

        thread_main = thread_main_factory()
        thread_main.start()

        thread_results = thread_results_factory()
        thread_results.start()

        while True:
            if not thread_update_connections.isAlive():
                logger.error("Thread thread_update_connections crashed in %s" % self.__class__.__name__)
                thread_update_connections = thread_update_connections_factory()
                thread_update_connections.start()
            if not thread_main.isAlive():
                logger.error("Thread thread_main crashed in %s" % self.__class__.__name__)
                thread_main = thread_results_factory()
                thread_main.start()
            if not thread_results.isAlive():
                logger.error("Thread thread_results crashed in %s" % self.__class__.__name__)
                thread_results = thread_results_factory()
                thread_results.start()
            time.sleep(10)

    def connection_thread(self, context=None, refresh_connection_stream=None, queue=None, process_id=None,
                          process_name=None):
        refresh_connection_sock = context.socket(zmq.SUB)
        refresh_connection_sock.connect(refresh_connection_stream)
        refresh_connection_sock.setsockopt(zmq.SUBSCRIBE, '')  # You must subscribe to something, so this means *all*
        set_timeouts_on_socket(refresh_connection_sock)

        # Wait for a result port and register as consumer of input stream
        while self.result_port is None:
            logger.debug("Waiting for result port")
            time.sleep(1)
        connections = get_connections_block('_update_connections', refresh_connection_sock)
        update_connection_sock = context.socket(zmq.PUSH)
        update_connection_sock.connect(connections['_update_connections']['streams'][0])
        update_connection_sock.send_json({
            'streams': {
                process_id: ['tcp://127.0.0.1:%s' % self.result_port]
            },
            'meta': {
                'id': process_id,
                'name': process_name,
                'grouping': None
            }
        })

        connections = get_connections_block('_message_ack', refresh_connection_sock)
        self.controller_sock = context.socket(zmq.PUSH)
        self.controller_sock.connect(connections['_message_ack']['streams'][0])
        set_timeouts_on_socket(self.controller_sock)

        while True:
            try:
                connections = refresh_connection_sock.recv_json()
                if queue in connections:
                    for send_conn in connections[queue]['streams']:
                        if send_conn not in self.send_socks:
                            send_sock = context.socket(zmq.PUSH)
                            send_sock.connect(send_conn)
                            self.send_socks[send_conn] = send_sock
                            self.send_grouper = connections[queue]['grouping']
            except zmq.Again:
                pass

    def message_producer(self, context=None, process_id=None):

        while True:
            if not self.send_socks:
                logger.debug("Waiting for send_socks")
                time.sleep(1)
            else:
                start_time = datetime.datetime.now()
                for received_message_result in self:
                    for generated_message in received_message_result:
                        if generated_message is not None:
                            try:
                                socket_address = self.get_grouper(self.send_grouper)(
                                    self.send_socks.keys()
                                ).get_destination_for(generated_message.grouping_value)
                            except GroupingValueMissing:
                                raise GroupingValueMissing("Message '%s' provided an invalid grouping_value: '%s'" % (generated_message.content, generated_message.grouping_value) )
                            generated_message.send(
                                self.send_socks[socket_address],
                                process_id
                            )
                            if self.controller_sock:
                                generated_message.send_control_message(
                                    self.controller_sock,
                                    time_consumed=datetime.datetime.now() - start_time,
                                    process_name=process_id,
                                    destination_endpoint=socket_address
                                )
                        start_time = datetime.datetime.now()

    def receive_replies(self, context=None):
        result_sock = context.socket(zmq.PULL)
        self.result_port = result_sock.bind_to_random_port("tcp://*")
        logger.debug("Result port is %s" % self.result_port)
        set_timeouts_on_socket(result_sock)

        while True:
            try:
                message_reply = result_sock.recv_json()
                if message_reply['status'] == 'success':
                    self.success(message_reply['id'])
                elif message_reply['status'] == 'fail':
                    self.failed(message_reply['id'])
                else:
                    logger.warn("Received unknown status feedback %s" % message_reply['status'])
            except zmq.Again:
                time.sleep(1)